import time, traceback, select
import random, queue
import argparse
from logging.handlers import RotatingFileHandler

from pyraft.common import *
from pyraft.protocol import resp
from pyraft.log import RaftLog
from pyraft.log import LogItem
from pyraft.worker.worker import MergedWorker
from pyraft.worker.redis_worker import RedisWorker
from pyraft.worker.base_worker import BaseWorker

class RaftNode(object):
	def __init__(self, nid, addr, ensemble={}, peer = False, worker = None, overwrite_peer=False):
		# raft node & peer common
		self.nid = nid
		self.term = 0
		self.index = 0
		self.state = 'c'
		self.last_append_entry_ts = 1
		self.last_delayed_ts = 1
		self.last_checkpoint = 0
		self.first_append_entry = False
		self.last_applied = 0
		self.commit_index = 0

		self.addr = addr
		self.ip, self.port = addr.split(':', 1)
		self.port = int(self.port)

		self.raft_req = resp.resp_io(None)
		self.raft_wait = resp.resp_io(None)

		if peer == True:
			return

		# raft node only
		self.logger = logging.getLogger('pyraft')
		self.shutdown_flag = False

		self.peers = {}
		self.peer_lock = threading.Lock()
		self.overwrite_peer = overwrite_peer

		self.log = RaftLog(nid)

		if worker is None:
			worker = MergedWorker(self.addr, BaseWorker(self.addr), RedisWorker(self.addr))

		self.worker = worker
		self.worker_map = {}
		self.worker_map[worker.worker_offset] = worker
		worker.init_node(self)

		self.data = {}
		self.data_lock = threading.Lock()
		self.data['ttl'] = {}
		self.ttl = self.data['ttl']

		for pid, paddr in ensemble.items():
			if pid == nid:
				continue

			if addr == paddr:
				continue

			self.add_node(pid, paddr)

	def regist_worker(self, worker_offset, worker):
		worker.worker_offset = worker_offset
		self.worker_map[worker_offset] = worker
		worker.init_node(self)

	def get_handler(self, name, worker_offset = 0):
		return self.worker_map[worker_offset].get_handler(name)

	def get_handler_func(self, name, worker_offset = 0): # return function only
		handler = self.get_handler(name, worker_offset)
		if isinstance(handler, list):
			return handler[0]

		return handler

	def propose(self, cmd, worker_offset=0, async_run=False):
		handler = self.get_handler(cmd[0].lower(), worker_offset)
		if handler is None:
			raise RaftException('unknown commands: %s' % cmd)

		if 'e' in handler[1]:
			if self.state == 'c':
				self.log_warn('request while candidate')
				raise RaftException('temporary unavailable')

			if self.state != 'l':
				for nid, p in self.get_peers().items():
					if p.state == 'l':
						return self.worker_map[worker_offset].relay_cmd(p, cmd, worker_offset)

				raise RaftException('cannot relay to leader')

			f = Future(cmd, worker_offset)
			self.q_entry.put(f)

			if async_run == True:
				return f

			ret = f.get(10)
			if ret == ERROR_APPEND_ENTRY:
				self.log_info('append_entry failed (%s)' % str(cmd))
		else:
			ret = handler[0](self, cmd)

		return ret

	def apply_loop(self):
		i = 0
		while True:
			if self.shutdown_flag:
				break

			if i % 10 == 0:
				# print self.get_snapshot()
				pass
			i += 1

			if self.log.size() > CONF_LOG_MAX:
				self.checkpoint()

			item = self.log.pop(1)
			if item == None:
				continue

			cmd = item.cmd
			worker_offset = item.worker_offset
			if isinstance(cmd, Future):
				cmd = cmd.cmd

			if self.index >= item.index:
				self.log_info('skip log [%d:%d]: "%s"' % (self.index, item.index, str(cmd)))
				continue

			self.log_debug('apply command [%d]: "%s"' % (item.index, str(cmd)))
			handler = self.get_handler(cmd[0].lower(), worker_offset)
			if handler is None:
				self.log_error('unknown command: %s' % cmd)
				sys.exit(-1)

			with self.data_lock:
				try:
					ret = handler[0](self, cmd)
				except RaftException as e:
					ret = e
				except Exception as e:
					print('unexpected exception: ', traceback.format_exc())
					ret = e

				self.index = item.index

			if isinstance(item.cmd, Future):
				item.cmd.set(ret)

	def load(self, filename):
		self.log_info('nid %s load %s' % (self.nid, filename))
		try:
			fh = open(filename, 'r')
			data = fh.read()
			fh.close()
		except IOError as e:
			self.log_error('failed to load: %s' % str(e))
			return False

		self.data = eval(data)
		meta = self.data['_META_']
		meta['id'] = self.nid
		meta['state'] = self.state
		self.term = meta['term']
		self.index = meta['index']
		self.log.index = self.index

		while True:
			try:
				fh = open('raft_%s_%010d.log' % (self.nid, self.index+1))
				remain = fh.read()
				fh.close()
			except IOError:
				break


			while True:
				l, remain = resp.resp_decoding(remain)
				if l == None:
					break

				# term, index, ts, worker_offset, cmd
				index = l[1]
				worker_offset = l[3]
				cmd = l[4]

				handler = self.get_handler(cmd[0].lower(), worker_offset)
				if handler is None:
					self.log_error('unknown command: %s' % cmd)
					sys.exit(-1)

				try:
					handler[0](self, cmd)
				except Exception:
					pass

				self.index = index
				self.log.index = self.index

				if remain == '':
					break

		return True

	def start(self):
		self.shutdown_flag = False

		self.q_entry = queue.Queue(4096)

		self.th_raft = threading.Thread(target = self.raft_listen)
		self.th_raft.start()

		self.th_le = threading.Thread(target = self.leader_election)
		self.th_le.start()

		self.th_apply = threading.Thread(target=self.apply_loop)
		self.th_apply.start()

		for offset in sorted(self.worker_map.keys()):
			worker = self.worker_map[offset]
			worker.start(self)

		self.on_start()

	def shutdown(self):
		for offset in sorted(self.worker_map.keys()):
			worker = self.worker_map[offset]
			worker.shutdown()

		self.shutdown_flag = True
		self.on_shutdown()

	def join(self):
		self.th_raft.join()
		self.th_le.join()
		self.th_apply.join()

		for offset in sorted(self.worker_map.keys()):
			worker = self.worker_map[offset]
			worker.join()

		for nid, peer in self.get_peers().items():
			peer.raft_req.close()
			peer.raft_wait.close()

		self.log.close()

	def add_node(self, nid, addr):
		with self.peer_lock:
			if nid == self.nid or nid in self.peers:
				self.log_warn('node %s already exists' % nid)
				return False

			if '__TEMP_%s__' % addr in self.peers: # replace temp peer
				del self.peers['__TEMP_%s__' % addr]

			for pid, peer in self.peers.items():
				if addr == peer.addr:
					self.log_warn('address %s already used in node %s' % (addr, pid))
					return False

			self.peers[nid] = RaftNode(nid, addr, peer = True)

		#self.raft_connect()
		return True

	def del_node(self, nid):
		with self.peer_lock:
			if nid not in self.peers:
				self.log_error('node %s not exists' % nid)
				return

			p = self.peers[nid]
			p.raft_req.close()
			p.raft_wait.close()
			del self.peers[nid]

	def get_peers(self):
		ret = {}
		with self.peer_lock:
			for nid, peer in self.peers.items():
				ret[nid] = peer
			
		return ret

	def raft_connect(self):
		for nid, peer in self.get_peers().items():
			if peer.raft_req.connected():
				continue

			try:
				sock = socket.socket()
				sock.connect((peer.ip, peer.port+1))
			except socket.error:
				sock.close()
				continue

			peer.raft_req = resp.resp_io(sock)
			self.log_info('connect to %s' % (nid))
			peer.raft_req.raw_write('id %s %s %d' % (self.nid, self.addr, self.index))

			peers = peer.raft_req.read(1)
			if not isinstance(peers, list):
				self.log_warn('connect to %s failed: "%s"' % (nid, str(peers)))
				return

			for p in peers:
				toks = p.split('/', 1)
				self.add_node(toks[0], toks[1])

			self.log_info('connect to %s ok' % nid)


	def process_raft_accept(self, sock):
		nid = None
		
		rio = resp.resp_io(sock)
		words = rio.read(1)
		if words == None or words == '':
			rio.close()
			return

		if isinstance(words, str):
			words = words.split()

		if len(words) == 4 and words[0] == 'id':
			nid = words[1]
			addr = words[2]
			index = intcast(words[3])
			if index == None:
				self.log_error('invalid id: %s', words)
				return

		self.log_info('raft accept: %s' % nid)

		if nid != None: # new peer
			if nid not in self.peers: # new node
				ret = self.add_node(nid, addr)
				if ret == False:
					rio.write(Exception('cannot add node (invalid nid or exists)'))
					rio.close()
					return
					
			peer = self.peers[nid]
			if peer.addr != addr:
				rio.write(Exception('nid already in ensemble'))
				rio.close()

				if self.overwrite_peer: # delete previous nid automatically (usually used in k8s environment. pod restart)
					self.del_node(nid)

				return
			else:
				# reconnect
				if peer.raft_wait != None:
					peer.raft_wait.close()

				peer.raft_wait = rio
				peers = ['%s/%s' % (self.nid, self.addr)]
				for nid, p in self.get_peers().items():
					peers.append('%s/%s' % (nid, p.addr))

				peer.raft_wait.write(peers)
				#self.log_info('peer write to %s' % peer.nid)
		else:
			self.log_error('invalid raft command: %s' % words)
			rio.write(Exception('invalid raft command'))
			rio.close()
			
	def raft_listen(self):
		self.raft_listen_sock = socket.socket()
		self.raft_listen_sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
		self.raft_listen_sock.bind((self.ip, self.port+1))
		self.raft_listen_sock.listen(1)
		self.raft_listen_sock.settimeout(1)

		while True:
			try:
				sock, addr = self.raft_listen_sock.accept()
				self.process_raft_accept(sock)
			except socket.timeout:
				if self.shutdown_flag:
					self.raft_listen_sock.close()
					break

	#
	# leader election
	#
	def leader_election(self):
		while True:
			self.raft_connect()

			if self.state == 'f':
				self.do_follower()
			elif self.state == 'c':
				self.do_candidate()
			elif self.state == 'l':
				self.do_leader()
			else:
				self.log_error('unknown state: %s' % self.state)

			if self.shutdown_flag:
				for nid, peer in self.get_peers().items():
					peer.raft_req.close()
				break

	def set_leader(self, node):
		if node.nid == self.nid:
			if self.state != 'l':
				self.first_append_entry = True
				self.on_leader()

			self.state = 'l'
		else:
			if self.state != 'f':
				self.on_follower()

			self.state = 'f'

		for nid, peer in self.get_peers().items():
			if node.nid == nid:
				peer.state = 'l'
			else:
				peer.state = 'f'

	def select_peer_req(self, timeout):
		sock_peer_map = {}
		for nid, p in self.get_peers().items():
			if p.raft_wait.sock != None:
				sock_peer_map[p.raft_wait.sock] = p

		if len(sock_peer_map) == 0:
			return []

		reads, writes, excepts = select.select(list(sock_peer_map.keys()), [], [], timeout)
		peers = []
		for r in reads:
			peers.append(sock_peer_map[r])
			
		return peers

	def handle_request(self, p, toks):
		#self.log_debug('handle req: %s' % str(toks))
		if toks[0] != 'append_entry' and toks[0] != 'snapshot':
			self.log_info('unknown or delayed request from %s: %s' % (p.nid, toks))
			return False

		term = intcast(toks[1])
		prev_term = intcast(toks[2])
		prev_index = intcast(toks[3])
		commit_index = intcast(toks[4])
		if term == None or prev_term == None or prev_index == None or commit_index == None:
			self.log_error('invalid append_entry: %s' % toks)
			return False

		if term < self.term:
			self.log_info('old term request from %s: %s' % (p.nid, toks))
			return False

		self.term = term
		self.set_leader(p)

		if self.commit_index != commit_index:
			self.commit_index = commit_index
			self.log.apply_commit_index(commit_index)

		if toks[0] == 'append_entry': # append_entry, term, prev_term, prev_index, commit_index, ts, worker_offset, cmds...
			ts = toks[5]
			if len(toks) > 6:
				self.log_debug('apply append_entry to %d-%d' % (term, prev_index))
				index = prev_index + 1
				item = LogItem(self.term, index, ts, int(toks[6]), toks[7:])
				self.log.push(item, self.commit_index)
			else:
				index = self.index
		elif toks[0] == 'snapshot': # snapshot, term, prev_term, prev_index, commit_index, data
			self.log_info('apply snapshot to %d-%d' % (term, prev_index))
			self.data = eval(toks[5])
			meta = self.data['_META_']
			meta['id'] = self.nid
			meta['state'] = self.state
			self.term = meta['term']
			self.index = meta['index']
			self.log.index = self.index
			index = self.index

		p.raft_wait.write('ack %d' % index)
		self.last_append_entry_ts = int(time.time())
		return True

	def handle_ack(self, p, expect = 0, timeout = 0.0):
		start = time.time()
		while True:
			now = time.time()
			if timeout > 0 and now - start > timeout:
				break
				
			msg_list = p.raft_req.read_all(0.0)
			if msg_list == None:
				return

			for toks in msg_list:
				if isinstance(toks, str):
					toks = toks.split()

				if toks[0] == 'ack':
					index = intcast(toks[1])
					if index == None:
						self.log_error('invalid ack: %s' + toks)

					p.state = 'f'
					p.term = self.term
					p.index = index
					p.last_append_entry_ts = time.time()
				else:
					self.log_info('unknown append_entry resp. from %s: "%s"' % (p.nid, toks))

			if p.index >= expect:
				break



	def do_follower(self):
		#self.log_info('do_follower')

		peers = self.select_peer_req(1.0)
		for p in peers:
			msg_list = p.raft_wait.read_all()
			if msg_list == None or msg_list == []:
				continue

			for toks in msg_list:
				if isinstance(toks, str):
					toks = toks.split()

				if toks[0] == 'vote':
					term = intcast(toks[1].strip())
					if term == None:
						self.log_error('invalid vote: %s' % toks)
						continue

					if term > self.term:
						p.raft_wait.write('yes')
					else:
						p.raft_wait.write('no')
				else:
					old_term = self.term
					self.handle_request(p, toks)
					if self.term > old_term:
						# split brain & new leader elected. 
						# clean data to install snapshot in case of async mode
						self.index = 0
						return

		if self.last_append_entry_ts > 0 and int(time.time()) - self.last_append_entry_ts > CONF_PING_TIMEOUT:
			self.on_candidate()
			self.state = 'c'


	def do_candidate(self):
		if len(self.get_peers()) > 0:
			connected = 0
			for nid, p in self.get_peers().items():
				if p.raft_req.connected():
					connected += 1
			if connected == 0:
				return

		#self.log_info('do_candidate')
		self.term += 1

		voting_wait = CONF_VOTING_TIME * 0.1
		vote_wait_timeout = random.randint(0, CONF_VOTING_TIME*1000  * 0.5) / 1000.0
		wait_remaining = 1 - vote_wait_timeout
		voted = False

		# process vote
		peers = self.select_peer_req(vote_wait_timeout)
		for p in peers:
			msg_list = p.raft_wait.read_all()
			if msg_list == None or msg_list == []:
				continue

			for toks in msg_list:
				if isinstance(toks, str):
					toks = toks.split()

				if toks[0] == 'vote':
					term = intcast(toks[1].strip())
					if term == None:
						self.log_error('invalid vote: %s' % toks)
						continue

					if not voted and term >= self.term:
						p.raft_wait.write('yes')
						voted = True
						self.term = term
					else:
						if term >= self.term:
							self.term = term
							
						p.raft_wait.write('no')
				else:
					if self.handle_request(p, toks):
						return # elected

		if voted:
			for nid, p in self.get_peers().items():
				msg_list = p.raft_wait.read_all(wait_remaining)
				if msg_list == None or msg_list == []:
					continue

				for toks in msg_list:
					if isinstance(toks, str):
						toks = toks.split()

					if self.handle_request(p, toks):
						return # elected

			return # not elected try next

		
		# process vote request
		count = 1
		voters = [self.nid]
		for nid, p in self.get_peers().items():
			p.raft_req.write('vote %d' % self.term)
		
		for i in range(2):
			get_result = {}
			for nid, p in self.get_peers().items():
				if nid in get_result:
					continue

				msg_list = p.raft_req.read_all(i*(CONF_VOTING_TIME/2))
				if msg_list == None or msg_list == []:
					continue

				for toks in msg_list:
					if isinstance(toks, str):
						toks = toks.split()

					if toks[0] == 'yes':
						voters.append(nid)
						count+=1
						get_result[nid] = True
					elif toks[0] == 'no':
						get_result[nid] = False
					else:
						self.handle_request(p, toks)

		# process result
		self.log_info('get %d. voters: %s' % (count, str(voters)))
		if count > (len(self.peers)+1)/2:
			self.log_info('%s is a leader' % (self.nid))
			self.set_leader(self)
			self.term += 10 


	def append_entry(self, future):
		ts = time.time()
		prev_index = self.log.get_index()
		prev_term = self.log.get_term()

		if future != None:
			append_cmd = ['append_entry', self.term, prev_term, prev_index, self.commit_index, ts]
			append_cmd.append(future.worker_offset)
			append_cmd += future.cmd
			for nid, p in self.get_peers().items():
				self.log_debug('leader write to %s: "%s"' % (p.nid, str(append_cmd)))
				p.raft_req.write(append_cmd)

			n_nodes = len(self.get_peers()) + 1
			half = n_nodes / 2.0
			n_ack = 1 # 1 for me

			for nid, p in self.get_peers().items():
				self.handle_ack(p, expect=prev_index+1, timeout=1.0)
				if p.index == prev_index+1:
					n_ack+=1

			if n_ack > half:
				self.commit_index = prev_index+1
				item = LogItem(self.term, prev_index+1, ts, future.worker_offset, future)
				self.log.push(item, self.commit_index)
				# send dummy append below to noti commit
			else:
				future.set(ERROR_APPEND_ENTRY)
				return

		append_cmd = ['append_entry', self.term, prev_term, prev_index, self.commit_index, ts]
		for nid, p in self.get_peers().items():
			p.raft_req.write(append_cmd)

	def get_pending_time(self): # get max diff ack time
		if self.state != 'l':
			return None # cannot determine

		now = time.time()
		max_diff = 0
		for nid, p in self.get_peers().items():
			if p.state == 'f':
				diff = now - p.last_append_entry_ts
				if diff > max_diff:
					max_diff = diff

		return max_diff

	def do_leader(self):
		#self.log_info('do_leader')
		for nid, p in self.get_peers().items():
			self.handle_ack(p)

		for nid, p in self.get_peers().items():
			now = time.time()
			if p.index == self.index:
				p.last_delayed_ts = now
				continue

			if now - p.last_delayed_ts > 2.0 and p.raft_req.connected() and p.index < self.index:
				p.last_delayed_ts = now
				self.process_install_snapshot(p)

		try:
			if self.first_append_entry:
				self.first_append_entry = False
				item = self.q_entry.get(False)
			else:
				item = self.q_entry.get(True, 1.0)
		except queue.Empty:
			item = None

		self.append_entry(item)

		# read peer request if exists
		peers = self.select_peer_req(0.0)
		for p in peers:
			msg_list = p.raft_wait.read_all()
			if msg_list == None or msg_list == []:
				continue

			for toks in msg_list:
				if isinstance(toks, str):
					toks = toks.split()

				if toks[0] == 'vote':
					p.raft_wait.write('no')
				else:
					old_term = self.term
					self.handle_request(p, toks)
					if self.term > old_term:
						# split brain & new leader elected. 
						# clean data to install snapshot in case of async mode
						self.index = 0
						return


	def get_snapshot(self):
		meta = {}
		meta['id'] = self.nid
		meta['term'] = self.term
		meta['index'] = self.index
		meta['state'] = self.state

		ensemble = {self.nid:self.addr}
		for nid, p in self.get_peers().items():
			ensemble[nid] = p.addr
		meta['ensemble'] = ensemble

		self.data['_META_'] = meta

		return self.data.__repr__()


	def checkpoint(self, filename=None):
		data = self.get_snapshot()
		flag_cleanup = False
		if filename == None:
			flag_cleanup = True
			filename = 'raft_%s_%d_%d.dat' % (self.nid, int(time.time()), self.index)

		fh = open(filename, 'w')
		fh.write(self.data.__repr__())
		fh.close()

		self.last_checkpoint = self.index
		
		if flag_cleanup:
			self.log.cleanup(self.index)

	def process_install_snapshot(self, p):
		diff = self.index - p.index
		prev_index = self.log.get_index()
		prev_term = self.log.get_term()


		if p.index < self.log.start_index() or (diff >= 100 or diff > len(self.data)/10):
			snapshot = self.get_snapshot()
			self.log_info('send snapshot to %s(%d)' % (p.nid, p.index))
			p.raft_req.write(['snapshot', self.term, prev_term, prev_index, self.commit_index, snapshot])
		else:
			old_logs = self.log.get_range(p.index) # term, index, ts, commands
			for l in old_logs:
				self.log_info('send append_entry to %s(%d)' % (p.nid, p.index))
				p.raft_req.write(['append_entry', self.term, l.term, l.index-1, self.commit_index, l.ts] + l.cmd)

	#
	# changed plugin. inherit or modify this (or add handler)
	#
	def on_start(self):
		self.log_info('on_start called')
		handler = self.get_handler_func('on_start')
		if handler is not None:
			handler(self)

	def on_shutdown(self):
		self.log_info('on_shutdown called')
		handler = self.get_handler_func('on_shutdown')
		if handler is not None:
			handler(self)

	def on_leader(self):
		self.log_info('on_leader called')
		handler = self.get_handler_func('on_leader')
		if handler is not None:
			handler(self)

	def on_follower(self):
		self.log_info('on_follower called')
		handler = self.get_handler_func('on_follower')
		if handler is not None:
			handler(self)

	def on_candidate(self):
		self.log_info('on_candidate called')
		handler = self.get_handler_func('on_candidate')
		if handler is not None:
			handler(self)
		
	#
	# log, etc
	#
	def log_debug(self, msg):
		log = '[%s-%d(%s)] %s' % (self.nid, self.term, self.state, msg)
		self.logger.debug(log)

	def log_info(self, msg):
		log = '[%s-%d(%s)] %s' % (self.nid, self.term, self.state, msg)
		self.logger.info(log)

	def log_warn(self, msg):
		log = '[%s-%d(%s)] %s' % (self.nid, self.term, self.state, msg)
		self.logger.warning(log)

	def log_error(self, msg):
		log = '[%s-%d(%s)] %s' % (self.nid, self.term, self.state, msg)
		self.logger.error(log)

	def check_ttl(self, key):
		if key in self.ttl:
			ttl = self.ttl[key]
			if ttl < time.time():
				del self.ttl[key]

				if key in self.data:
					del self.data[key]

	def clear_ttl(self, key):
		if key in self.ttl:
			del self.ttl[key]
			
	def set_ttl(self, key, ts):
		if key in self.data:
			self.ttl[key] = ts
			return True
		else:
			return False

	def request(self, *cmd):
		try:
			ret = self.propose(cmd)
		except RaftException as e:
			ret = e
		except Exception as e:
			print('unexpected exception: ', traceback.format_exc())
			ret = e

		return ret

	def request_async(self, *cmd):
		return self.propose(cmd, async_run=True)

def parse_default_args(parser):
	parser.add_argument('-a', dest='addr', help='ip:port[port+1], :port means pick one ip by gethostbyname (ex. -a 127.0.0.1:5010)')
	parser.add_argument('-e', dest='ensemble', help='ensemble ip list or domain name with port (ex. -e 2/127.0.0.1:5020,3/127.0.0.1:5030 or -e 127.0.0.1:5020,127.0.0.1:5030 or -e pyraft.test.com:5010)')
	parser.add_argument('-i', dest='nid', help='self node id (if not exists, use address, HOSTNAME use machine hostname) (ex. -i 1)')
	parser.add_argument('-l', dest='load', help='checkpoint filename to load')
	parser.add_argument('-o', dest='overwrite_peer', help='overwrite duplicated nid node (delete previous one)', action='store_true')
	parser.add_argument('-loglevel', dest='loglevel', default='info', help='loglevel (debug, info, warning, error, fatal)')
	parser.add_argument('-logfile', dest='logfile', help='logger rotation file')

	args = parser.parse_args()

	## process log level & log file
	if args.loglevel.lower() != 'warning':
		logger = logging.getLogger('pyraft')

		if args.loglevel.lower() == 'debug':
			logger.setLevel(logging.DEBUG)
		elif args.loglevel.lower() == 'info':
			logger.setLevel(logging.INFO)
		elif args.loglevel.lower() == 'error':
			logger.setLevel(logging.ERROR)
		elif args.loglevel.lower() == 'fatal':
			logger.setLevel(logging.FATAL)
		else:
			raise RaftException('unknown log level')

	if args.logfile is not None:
		handler = RotatingFileHandler(args.logfile, maxBytes=1024*1024, backupCount=10)
		formatter = logging.Formatter('%(asctime)s [%(levelname)s] %(message)s')
		handler.setFormatter(formatter)
		logger.addHandler(handler)

	## process ensemble
	if args.addr == None:
		parser.print_help()
		raise RaftException('addr is required')

	if args.addr.startswith(':'):
		ip = socket.gethostbyname(socket.gethostname())
		args.addr = '%s%s' % (ip, args.addr)

	if args.nid == None:
		args.nid = args.addr

	if args.nid == 'HOSTNAME':
		args.nid = socket.gethostname()

	ensemble = {}
	if args.ensemble != None:
		is_domain_name = False
		for c in args.ensemble:
			if c.isalpha():
				is_domain_name = True
				break

		if is_domain_name:
			if ':' not in args.ensemble:
				print('domain name ensemble should include port')
				sys.exit(-1)

			domain_name, port = args.ensemble.split(':', 1)
			try:
				host, alias, ip_list = socket.gethostbyname_ex(domain_name)
				for ip in ip_list:
					addr = '%s:%d' % (ip, int(port))
					ensemble['__TEMP_%s__' % addr] = addr
			except socket.gaierror: # in k8s DNS is setup later
				pass
		else:
			toks = args.ensemble.split(',')
			for tok in toks:
				etoks = tok.split('/')
				if len(etoks) == 2:
					nid = etoks[0]
					addr = etoks[1]
					ensemble[nid] = addr
				elif len(etoks) == 1:
					addr = tok
					if addr.startswith(':'):
						ip = socket.gethostbyname(socket.gethostname())
						addr = '%s%s' % (ip, addr)

					ensemble['__TEMP_%s__' % addr] = addr
				else:
					print('invalid ensemble format')
					sys.exit(-1)

	#print(ensemble)
	args.ensemble_map = ensemble
	return args

def make_default_node(): # redis interface node is default now
	args = parse_default_args(argparse.ArgumentParser())
	node = RaftNode(args.nid, args.addr, args.ensemble_map, overwrite_peer=args.overwrite_peer)

	if args.load != None:
		node.load(args.load)

	return node





import json

from pyraft.common import *
from pyraft.protocol.base import base_io

def resp_encoding(msg):
	if isinstance(msg, bool) and msg == True:
		return '+OK\r\n'

	if msg == None:
		return '$-1\r\n'

	if isinstance(msg, str):
		if len(msg) < 10:
			return '+%s\r\n' % msg
		else:
			return '$%d\r\n%s\r\n' % (len(msg), msg)

	if isinstance(msg, int):
		return ':%d\r\n' % msg

	if isinstance(msg, float): 
		return '+%s\r\n' % str(msg)

	if isinstance(msg, list) or isinstance(msg, tuple):
		result = ['*%d\r\n' % len(msg)]
		for item in msg:
			result.append(resp_encoding(item))

		return ''.join(result)
			
	if isinstance(msg, Exception):
		return '-%s\r\n' % str(msg)

	# json encodable class (ex. zk Commands classes)
	if hasattr(msg, 'json'):
		result = {}
		for key, value in msg.__dict__.items():
			if key.startswith('_'):
				continue
			result[key] = value

		return resp_encoding(json.dumps(result))

	return '-unknown resp type\r\n'

def resp_decoding(src):
	#print('decoding >>>> "%s"' % src)
	if len(src) == 0:
		return None, src

	if src.startswith(b'*'):
		toks = src.split(b'\r\n', 1)
		count = int(toks[0][1:])
		remain = toks[1]

		result = []
		for i in range(count):
			item, remain = resp_decoding(remain)
			if item == None:
				return None, src

			result.append(item)

		return result, remain

	if src.startswith(b'$'):
		toks = src.split(b'\r\n', 1)
		if len(toks) != 2:
			return None, src

		size = int(toks[0][1:])
		remain = toks[1]
		if len(remain) < size+2:
			return None, src

		return remain[:size].decode('utf-8'), remain[size+2:]
		
	if src.startswith(b'+'):
		toks = src.split(b'\r\n', 1)
		if len(toks) != 2:
			return None, src

		return toks[0][1:].decode('utf-8'), toks[1]

	if src.startswith(b'-'):
		toks = src.split(b'\r\n', 1)
		if len(toks) != 2:
			return None, src

		return Exception(toks[0][1:].decode('utf-8')), toks[1]

	if src.startswith(b':'):
		toks = src.split(b'\r\n', 1)
		if len(toks) != 2:
			return None, src

		return int(toks[0][1:]), toks[1]

	# plain text input
	toks = src.split(b'\r\n', 1)
	if len(toks) == 1:
		return src.decode('utf-8'), b''

	return toks[0].decode('utf-8'), toks[1]


class resp_io(base_io):
	def __init__(self, sock):
		super().__init__(sock)

	def raw_encode(self, msg):
		if not msg.endswith('\r\n'):
			msg += '\r\n'

		return msg

	def encode(self, msg):
		return resp_encoding(msg)

	def decode(self, msg):
		return resp_decoding(msg)

	def decodable(self, buff):
		if b'\r\n' in buff:
			return True

		return False



import struct

from pyraft.common import *
from pyraft.protocol.base import base_io
from pyraft.protocol.zk_exceptions import *

# came from kazoo serialization.py
# Struct objects with formats compiled
bool_struct = struct.Struct('B')
int_struct = struct.Struct('!i')
int_int_struct = struct.Struct('!ii')
int_int_long_struct = struct.Struct('!iiq')

int_long_int_long_struct = struct.Struct('!iqiq')
long_struct = struct.Struct('!q')
multiheader_struct = struct.Struct('!iBi')
reply_header_struct = struct.Struct('!iqi')
stat_struct = struct.Struct('!qqqqiiiqiiq')

session_map = {}
def get_session_io(session_id):
    return session_map.get(session_id)

def set_session_io(session_id, io):
    global session_map
    session_map[session_id] = io

def del_session_io(session_id):
    global session_map
    if session_id in session_map:
        del session_map[session_id]

def read_string(buffer, offset):
    """Reads an int specified buffer into a string and returns the
    string and the new offset in the buffer"""
    length = int_struct.unpack_from(buffer, offset)[0]
    offset += int_struct.size
    if length < 0:
        return None, offset
    else:
        index = offset
        offset += length
        return buffer[index:index + length].decode('utf-8'), offset

def read_acl(bytes, offset):
    perms = int_struct.unpack_from(bytes, offset)[0]
    offset += int_struct.size
    scheme, offset = read_string(bytes, offset)
    id, offset = read_string(bytes, offset)
    return (perms, scheme, id), offset

def write_string(bytes):
    if not bytes:
        return int_struct.pack(-1)
    else:
        utf8_str = bytes.encode('utf-8')
        return int_struct.pack(len(utf8_str)) + utf8_str

def write_buffer(bytes):
    if bytes is None:
        return int_struct.pack(-1)
    else:
        return int_struct.pack(len(bytes)) + bytes

def read_buffer(bytes, offset):
    length = int_struct.unpack_from(bytes, offset)[0]
    offset += int_struct.size
    if length < 0:
        return None, offset
    else:
        index = offset
        offset += length
        return bytes[index:index + length], offset

class ZkWatch:
    def __init__(self, type, state, path):
        self.type = type
        self.state = state
        self.path = path

    def serialize(self, b):
        b.extend(int_struct.pack(self.type))
        b.extend(int_struct.pack(self.state))
        b.extend(write_string(self.path))
        return b

class ZkClose:
    type = -11

    def deserialize(self, buff, offset):
        return ['close', self]

    def serialize(self, b):
        return b

class ZkConnect:
    type = None

    def __init__(self):
        self.connected = False

    def deserialize(self, buff, offset):
        self.proto_ver = int_struct.unpack_from(buff, offset)
        self.proto_ver, last_zxid, self.timeout, self.session_id = int_long_int_long_struct.unpack_from(buff, offset)
        offset += int_long_int_long_struct.size

        self.read_only = 0
        if self.proto_ver != -1: # relay connection skip password, read_only
            password, offset = read_buffer(buff, offset)
            self.read_only = buff[offset]

        self.connected = True
        return ['connect', self]

    def serialize(self, b):
        b.extend(int_int_long_struct.pack(self.proto_ver, self.timeout, self.session_id))
        b.extend(write_buffer(self.password))
        b.extend([self.read_only])
        return b

class ZkCreate:
    type = 1
    write = True # save original buffer & request socket to handle relay
    json = True # convert __dict__ to json for logging

    def deserialize(self, buff, offset):
        self.path, offset = read_string(buff, offset)
        data, offset = read_buffer(buff, offset)
        if data is None:
            self.data = ''
        else:
            self.data = data.decode('utf-8')

        acl_len = int_struct.unpack_from(buff, offset)[0]
        offset += int_struct.size

        self.acl = []
        for i in range(acl_len):
            perms = int_struct.unpack_from(buff, offset)[0]
            offset += int_struct.size
            id_scheme, offset = read_string(buff, offset)
            id_id, offset = read_string(buff, offset)
            self.acl.append([perms, id_scheme, id_id])

        self.flags = int_struct.unpack_from(buff, offset)[0]
        offset += int_struct.size

        return ['create', self]

    def serialize(self, b):
        b.extend(write_string(self.path))
        return b

class ZkDelete:
    type = 2
    write = True
    json = True

    def deserialize(self, buff, offset):
        self.path, offset = read_string(buff, offset)
        self.version = int_struct.unpack_from(buff, offset)[0]
        offset += int_struct.size
        return ['delete', self]

    def serialize(self, b):
        return b

class ZkExists:
    type = 3

    def deserialize(self, buff, offset):
        self.path, offset = read_string(buff, offset)
        self.watcher = (buff[offset] == 1)
        return ['exists', self]

    def serialize(self, b):
        b.extend(self._node.stat_pack())
        return b

class ZkGetData:
    type = 4

    def deserialize(self, buff, offset):
        self.path, offset = read_string(buff, offset)
        self.watcher = (buff[offset] == 1)
        return ['getdata', self]

    def serialize(self, b):
        b.extend(write_string(self._node.get_data()))
        b.extend(self._node.stat_pack())
        return b

class ZkSetData:
    type = 5
    write = True
    json = True

    def deserialize(self, buff, offset):
        self.path, offset = read_string(buff, offset)
        data, offset = read_buffer(buff, offset)
        if data is None:
            self.data = ''
        else:
            self.data = data.decode('utf-8')

        self.version = int_struct.unpack_from(buff, offset)[0]
        offset += int_struct.size
        return ['setdata', self]

    def serialize(self, b):
        b.extend(self._node.stat_pack())
        return b

class ZkGetACL:
    type = 6

    def deserialize(self, buff, offset):
        self.path, offset = read_string(buff, offset)
        return ['getacl', self]

    def serialize(self, b):
        count = len(self.acl)

        b.extend(int_struct.pack(count))
        if count <= 0:
            return b

        for acl in self.acl:
            b.extend(int_struct.pack(acl[0]))   # perms
            b.extend(write_string(acl[1]))      # scheme
            b.extend(write_string(acl[2]))      # id

        return b

class ZkSetACL:
    type = 7
    write = True
    json = True

    def deserialize(self, buff, offset):
        self.path, offset = read_string(buff, offset)
        acl_len = int_struct.unpack_from(buff, offset)[0]
        offset += int_struct.size

        self.acl = []
        for i in range(acl_len):
            perms = int_struct.unpack_from(buff, offset)
            offset += int_struct.size
            id_scheme, offset = read_string(buff, offset)
            id_id, offset = read_string(buff, offset)
            self.acl.append((perms, id_scheme, id_id))

        return ['setacl', self]

    def serialize(self, b):
        b.extend(self._node.stat_pack())
        return b

class ZkGetChildren:
    type = 8

    def deserialize(self, buff, offset):
        self.path, offset = read_string(buff, offset)
        self.watcher = (buff[offset] == 1)
        return ['getchildren', self]

    def serialize(self, b):
        count = len(self._children)

        b.extend(int_struct.pack(count))
        for child in self._children:
            b.extend(write_string(child.name))

        if self.type == 12: # getchildren2
            b.extend(self._node.stat_pack())

        return b

class ZkGetChildren2(ZkGetChildren):
    type = 12

class ZkPing:
    type = 11

    def deserialize(self, buff, offset):
        return ['ping', self]

    def serialize(self, b):
        return b

class zk_io(base_io):
    request_map = {
        -11:ZkClose,
        1:ZkCreate,
        2:ZkDelete,
        3:ZkExists,
        4:ZkGetData,
        5:ZkSetData,
        6:ZkGetACL,
        7:ZkSetACL,
        8:ZkGetChildren,
        11:ZkPing,
        12:ZkGetChildren2,
    }

    zxid = 0

    def __init__(self, sock):
        super().__init__(sock)
        self.conn = ZkConnect()
        self.xid = 0
        self.relay_map = {}

    def inc_zxid(self):
        self.zxid += 1
        return self.zxid

    def raw_encode(self, buff):
        return buff

    def encode(self, cmd):
        #print('>> encoding: %s' % str(cmd))
        if isinstance(cmd, bytes) or isinstance(cmd, bytearray):
            return cmd

        if isinstance(cmd, str): # like 4 letter command
            return cmd.encode()

        b = bytearray()
        # make header (if not connect)

        if isinstance(cmd, ZkConnect):
            logger.info("get new session: %s" % cmd.session_id)
            set_session_io(cmd.session_id, self)
        else:
            error_code = 0
            xid = self.xid
            if isinstance(cmd, ZookeeperError):
                error_code = cmd.code
            if isinstance(cmd, ZkPing):
                zxid = self.zxid
            if isinstance(cmd, ZkWatch):
                xid = -1
                zxid = -1
            else:
                zxid = self.inc_zxid()

            b.extend(reply_header_struct.pack(xid, zxid, error_code))

        # make body if not error
        if not isinstance(cmd, ZookeeperError):
            b = cmd.serialize(b)

        b = int_struct.pack(len(b)) + b
        return b

    def decode(self, buff):
        length = int_struct.unpack_from(buff, 0)[0]
        remain = buff[int_struct.size + length:]
        offset = int_struct.size

        if self.conn.connected == False:
            return self.conn.deserialize(buff, offset), remain

        self.xid, request_type = int_int_struct.unpack_from(buff, offset)
        offset += int_int_struct.size

        #print("decode buff: ",  self.sock.fileno(), bytes_to_str(buff))
        if request_type not in self.request_map:
            raise UnimplementedError()

        cmd = self.request_map[request_type]()
        cmd.xid = self.xid
        cmd.session_id = self.conn.session_id

        if hasattr(cmd, 'write'):
            cmd._original_buff = buff[:int_struct.size + length]
            cmd._req_io = self

        return cmd.deserialize(buff, offset), remain

    def decodable(self, buff):
        if len(buff) < 4:
            return False

        length = int_struct.unpack_from(buff, 0)[0]
        if len(buff) > (length + 4):
            return True

        return False

    # TODO: enhance
    def relay_cmd(self, addr, cmd): # cmd: ['create', ZkCreate]
        name = cmd[0]
        cmd = cmd[1]

        if addr not in self.relay_map:
            ip, port = addr.split(':')
            s = socket.socket()
            s.connect((ip, int(port)))
            self.relay_map[addr] = s

            # make relay connection
            b = bytearray()
            b.extend(int_struct.pack(int_long_int_long_struct.size))
            b.extend(int_long_int_long_struct.pack(-1, 0, 10000, cmd.session_id)) # proto, last zxid, timeout, session_id
            s.send(b)
            ret = s.recv(4096)

        s = self.relay_map[addr]
        s.send(cmd._original_buff)
        ret = s.recv(4096)

        return ret

    def close(self):
        super().close()
        del_session_io(self.conn.session_id)



import json
import random
import struct

from pyraft.common import *
from pyraft.worker.worker import Worker
from pyraft.protocol import zk
from pyraft.protocol.zk_exceptions import *
from pyraft.worker.zk_watcher import ZkWatcher
from pyraft.worker.zk_ephemeral import ZkEphermeralManager

stat_struct = struct.Struct('!qqqqiiiqiiq')

class ZkProtocol:
    def open_io(self, handle):
        return zk.zk_io(handle)

# TODO: r/w locking
def init_zk_stat():
    stat = {}
    stat['ctime'] = stat['mtime'] = int(time.time() * 1000)
    stat['czxid'] = stat['mzxid'] = 0
    stat['version'] = 0
    stat['cversion'] = 0
    stat['aversion'] = 0
    stat['ephermeralOwner'] = 0
    stat['pzxid'] = 0

    return stat

class ZkNode(object):
    def __init__(self, name, obj):
        self.name = name
        self.obj = obj

    def __getattr__(self, name):
        if name in self.obj:
            return self.obj[name]

        raise Exception('%s not found' % name)

    def __setattr__(self, key, value):
        if key in ['name', 'obj']:
            super().__setattr__(key, value)
        else:
            self.obj[key] = value

    def is_ephemeral(self):
        return (self.flags & 1) != 0

    def is_sequence(self):
        return (self.flags & 2) != 0

    def stat_pack(self):
        s = self.stat
        n_child = 0
        for v in self.obj.values():
            if isinstance(v, dict):
                n_child += 1

        return stat_struct.pack(s['czxid'], s['mzxid'], s['ctime'], s['mtime'],
                                s['version'], s['cversion'], s['aversion'],
                                s['ephermeralOwner'], len(self.data), n_child,
                                s['pzxid'])

    def get_data(self):
        return self.data

    def set_data(self, data):
        self.data = data
        self.stat['version'] += 1

    def get_acl(self):
        return self.acl

    def set_acl(self, acl):
        self.acl = acl
        self.stat['aversion'] += 1

    def get_flags(self):
        return self.flags

    def set_flags(self, flags):
        self.flags = flags

    def get_child(self, name):
        if name not in self.obj:
            raise NoNodeError()

        if not isinstance(self.obj[name], dict):
            raise NoNodeError()

        return ZkNode(name, self.obj[name])

    def create_child(self, name, data, acl, flags):
        if name in ['data', 'acl', 'flags', 'stat']:
            logger.error('%s is not allowed in zk tree' % name)
            raise BadArgumentsError()

        if name in self.obj:
            raise NodeExistsError()

        self.obj[name] = {'data':data, 'acl':acl, 'flags':flags, 'stat':init_zk_stat()}
        self.stat['cversion'] += 1
        return ZkNode(name, self.obj[name])

    def delete_child(self, name):
        if name not in self.obj:
            raise NoNodeError()

        child = self.obj[name]

        for k, v in child.items():
            if k != 'stat' and isinstance(v, dict):
                raise NotEmptyError()

        del self.obj[name]
        self.stat['cversion'] += 1

    def get_children(self):
        ret = []
        for k, v in self.obj.items():
            if k != 'stat' and isinstance(v, dict):
                ret.append(ZkNode(k, v))

        return ret

# handle string based request (by append_entry)
def handle_json(handler):
    convert_table = {
        'create':zk.ZkCreate,
        'delete':zk.ZkDelete,
        'setdata':zk.ZkSetData,
        'setacl':zk.ZkSetACL
    }

    def convert_to_command(*args, **kwargs): # self, node, words (str, cmd)
        words = args[2]
        if not isinstance(words[1], str):
            return handler(*args, **kwargs)

        cmd = convert_table[words[0]]()
        obj = json.loads(words[1])
        cmd.__dict__.update(obj)

        words[1] = cmd
        return handler(*args, **kwargs)

    return convert_to_command

class ZkWorker(Worker):
    def __init__(self, addr):
        super().__init__(addr)
        self.init_zk_handler()
        self.set_protocol(ZkProtocol())
        self.session_map = {} # connected sessions

    def start(self, node):
        super().start(node)
        self.ephemeral_mgr = ZkEphermeralManager(node)
        self.ephemeral_mgr.start()
        self.watch_mgr = ZkWatcher(node)

    def shutdown(self):
        super().shutdown()
        self.ephemeral_mgr.quit_flag = True

    def join(self):
        super().join()
        self.ephemeral_mgr.join()

    def init_node(self, node):
        if 'ZK' not in node.data:
            node.data['ZK'] = {'data':'', 'acl':[], 'flags':0, 'stat':init_zk_stat()}

        self.node = node

    def init_zk_handler(self):
        self.handler['connect'] = [self.do_connect, 'r', 1, 1]
        self.handler['create'] = [self.do_create, 'we', 1, 1]
        self.handler['create2'] = [self.do_create, 'we', 1, 1]
        self.handler['exists'] = [self.do_exists, 'r', 1, 1]
        self.handler['delete'] = [self.do_delete, 'we', 1, 1]
        self.handler['getdata'] = [self.do_get_data, 'r', 1, 1]
        self.handler['setdata'] = [self.do_set_data, 'we', 1, 1]
        self.handler['getacl'] = [self.do_get_acl, 'r', 1, 1]
        self.handler['setacl'] = [self.do_set_acl, 'we', 1, 1]
        self.handler['getchildren'] = [self.do_get_children, 'r', 1, 1]
        self.handler['getchildren2'] = [self.do_get_children, 'r', 1, 1]
        self.handler['close'] = [self.do_close, 'r', 1, 1]
        self.handler['ping'] = [self.do_ping, 'r', 1, 1]
        '''
        self.handler['sync'] = [self.do_sync, 'we', 1, 1]
        self.handler['reconfig'] = [self.do_reconfig, 'we', 1, 1]
        self.handler['sasl'] = [self.do_sasl, 'r', 1, 1]
        self.handler['auth'] = [self.do_auth, 'r', 1, 1]
        self.handler['transaction'] = [self.do_transaction, 'r', 1, 1]
        self.handler['checkversion'] = [self.do_checkversion, 'r', 1, 1]
        '''

    def do_connect(self, node, words):
        cmd = words[1]
        if cmd.session_id == 0:
            cmd.session_id = random.getrandbits(63)

        cmd.password = bytearray([random.getrandbits(8) for i in range(16)])

        node.request('hset', 'zk_session', str(cmd.session_id), str(int(time.time())))
        return cmd

    def do_close(self, node, words):
        cmd = words[1]

        node.request('hdel', 'zk_session', str(cmd.session_id))
        return {'quit_after_send':cmd}

    def _cd_path(self, node, path, parent=False):
        if path[0] != '/':
            raise BadArgumentsError()

        if len(path) > 1 and path[-1] == '/':
            raise BadArgumentsError()

        if path.strip() == '/':
            return ZkNode('ZK', node.data['ZK'])

        dir_list = path[1:].split('/')
        cwd = ZkNode('ZK', node.data['ZK'])
        count = len(dir_list)
        if parent:
            count -= 1

        for i in range(count):
            dir = dir_list[i]
            if dir == '':
                raise BadArgumentsError()

            dir = dir_list[i]
            cwd = cwd.get_child(dir)

        return cwd

    @handle_json
    def do_create(self, node, words): # create, create2
        cmd = words[1]
        basename = cmd.path.split('/')[-1]
        cwd = self._cd_path(node, cmd.path, parent=True)
        cmd._child = cwd.create_child(basename, cmd.data, cmd.acl, cmd.flags)

        self.watch_mgr.check_data_watch(cmd.path, ZkWatcher.EVENT_CREATED)
        self.watch_mgr.check_child_watch(cmd.path)

        if cmd._child.is_ephemeral():
            cmd._child.stat['ephemeralOwner'] = cmd.session_id
            self.ephemeral_mgr.regist(cmd.path, cmd._child)

        return cmd

    @handle_json
    def do_delete(self, node, words):
        cmd = words[1]
        basename = cmd.path.split('/')[-1]

        cwd = self._cd_path(node, cmd.path)
        if cwd.is_ephemeral():
            self.ephemeral_mgr.unregist(cmd.path)

        cwd = self._cd_path(node, cmd.path, parent=True)
        cwd.delete_child(basename)

        parent_path = '/'.join(cmd.path.split('/')[:-1]) # '/aaa/bbb' -> '/aaa'
        if parent_path == '':
            parent_path = '/'

        self.watch_mgr.check_data_watch(cmd.path, ZkWatcher.EVENT_DELETED)
        self.watch_mgr.check_child_watch(parent_path)

        return cmd

    def do_exists(self, node, words):
        cmd = words[1]
        if cmd.watcher:
            self.watch_mgr.regist_data_watch(cmd.path, cmd.session_id)

        cmd._node = self._cd_path(node, cmd.path)
        return cmd

    def do_get_data(self, node, words):
        cmd = words[1]
        cwd = self._cd_path(node, cmd.path)
        cmd._node = cwd
        if cmd.watcher:
            self.watch_mgr.regist_data_watch(cmd.path, cmd.session_id)

        return cmd

    @handle_json
    def do_set_data(self, node, words):
        cmd = words[1]
        cwd = self._cd_path(node, cmd.path)
        cwd.set_data(cmd.data)
        cmd._node = cwd

        self.watch_mgr.check_data_watch(cmd.path, ZkWatcher.EVENT_CHANGED)
        return cmd

    def do_get_acl(self, node, words):
        cmd = words[1]
        cwd = self._cd_path(node, cmd.path)
        cmd.acl = cwd.get_acl()
        return cmd

    @handle_json
    def do_set_acl(self, node, words):
        cmd = words[1]
        cwd = self._cd_path(node, cmd.path)
        cwd.set_acl(cmd.acl)
        cmd._node = cwd
        return cmd

    def do_get_children(self, node, words): # get_children, get_children2
        cmd = words[1]
        cwd = self._cd_path(node, cmd.path)
        cmd._children = cwd.get_children()
        cmd._node = cwd

        if cmd.watcher:
            self.watch_mgr.regist_child_watch(cmd.path, cmd.session_id)

        return cmd

    def do_ping(self, node, words):
        cmd = words[1]
        logger.debug('process ping session: %s' % cmd.session_id)
        node.request('hset', 'zk_session', str(cmd.session_id), str(int(time.time())))
        return cmd

    def relay_cmd(self, leader, cmd, worker_offset): # cmd: ['create', ZkCreate]
        p = leader
        try:
            ip, port = p.addr.split(':')
            new_addr = '%s:%d' % (ip, int(port)+worker_offset)
            ret = cmd[1]._req_io.relay_cmd(new_addr, cmd)
            return {'bypass':ret}

        except Exception as e:
            raise RaftException('relay to leader has exception: %s' % str(e))

    # TODO: transaction
    # TODO: checkversion
    # TODO: sync
    # TODO: reconfig
    # TODO: auth
    # TODO: sasl
    # TODO: sequence



import logging

from pyraft import raft
from pyraft.protocol import resp

__all__ = ['raft', 'resp']
__version__ = '0.2.0'

logger = logging.getLogger('pyraft')

handler = logging.StreamHandler()
formatter = logging.Formatter('%(asctime)s [%(levelname)s] %(message)s')
handler.setFormatter(formatter)
logger.addHandler(handler)

logger.setLevel(logging.WARN)




from pyraft.common import *
from pyraft.protocol import zk

class ZkWatcher:
    EVENT_NONE = -1
    EVENT_CREATED = 1
    EVENT_DELETED = 2
    EVENT_CHANGED = 3
    EVENT_CHILD = 4

    STATE_UNKNOWN = -1
    STATE_DISCONNECTED = 0
    STATE_NO_SYNC_CONNECTED = 1 # deprecated
    STATE_SYNC_CONNECTED = 3
    STATE_AUTH_FAILED = 4
    STATE_CONNECTED_READ_ONLY = 5
    STATE_SASL_AUTHENTICATED = 6
    STATE_CLOSED = 7
    STATE_EXPIRED = -112

    def __init__(self, node):
        self.lock = threading.Lock()
        self.node = node

    def send_watch_notification(self, session, event, path):
        print('>> send watch notification to', session, event, path)
        io = zk.get_session_io(session)
        if io is None:
            print('no session for watch')
            #TODO: warning
            return

        noti = zk.ZkWatch(event, self.STATE_SYNC_CONNECTED, path)
        io.write(noti)

    def check_child_watch(self, path):
        with self.lock:
            session_list = self.node.data.get('zk_watch_child_%s' % path)
            if session_list == None:
                return

            for session in session_list:
                self.send_watch_notification(session, self.EVENT_CHILD, path)

            del self.node.data['zk_watch_child_%s' % path]

        return []

    def check_data_watch(self, path, event):
        with self.lock:
            session_list = self.node.data.get('zk_watch_data_%s' % path)
            if session_list == None:
                return

            for session in session_list:
                self.send_watch_notification(session, event, path)

            del self.node.data['zk_watch_data_%s' % path]

        return []

    def regist_child_watch(self, path, session_id):
        with self.lock:
            self.node.request_async('rpush', 'zk_watch_child_%s' % path, session_id)

    def regist_data_watch(self, path, session_id):
        with self.lock:
            self.node.request_async('rpush', 'zk_watch_data_%s' % path, session_id)



from pyraft.common import *
from pyraft.worker.worker import Worker
from pyraft.worker.worker import RespProtocol


# worker for basic control (info, shutdown, add_node, del_node etc)
class BaseWorker(Worker):
    def __init__(self, addr):
        super().__init__(addr)
        self.init_base_handler()
        self.set_protocol(RespProtocol())

    # inherit & extend this interface
    def init_base_handler(self):
        self.handler['info'] = [self.do_info, 'r', 0, 0]
        self.handler['shutdown'] = [self.do_shutdown, 'e', 0, 0]
        self.handler['quit'] = [self.do_quit, 'r', 0, 0]

        self.handler['add_node'] = [self.do_add_node, 'we', 2, 2]
        self.handler['del_node'] = [self.do_del_node, 'we', 1, 1]
        self.handler['checkpoint'] = [self.do_checkpoint, 'r', 0, 0]
        self.handler['getdump'] = [self.do_getdump, 'r', 0, 0]
        self.handler['getlog'] = [self.do_getlog, 'r', 0, 2]

    def do_info(self, node, words):
        peers = {}
        for nid, p in node.get_peers().items():
            peers[nid] = {'state': p.state, 'addr': p.addr, 'term': p.term, 'index': p.index}

        info = {'nid': node.nid, 'state': node.state, 'term': node.term, 'index': node.index,
                'peers': peers, 'last_checkpoint': node.last_checkpoint}
        return str(info).replace("'", '"')

    def do_quit(self, node, words):
        return {'quit': True}

    def do_shutdown(self, node, words):
        node.shutdown()
        return True

    def do_add_node(self, node, words):
        node.add_node(words[1], words[2])
        return True

    def do_del_node(self, node, words):
        node.del_node(words[1])
        return True

    def do_checkpoint(self, node, words):
        name = None
        if len(words) > 1:
            name = words[1]

        node.checkpoint(name)
        return True

    def do_getdump(self, node, words):
        return node.get_snapshot()

    def do_getlog(self, node, words):
        start = 0
        end = -1

        if len(words) > 1:
            start = intcast(words[1])
            if start == None:
                raise ERROR_CAST

        if len(words) > 2:
            end = intcast(words[2])
            if end == None:
                raise ERROR_CAST

        result = node.log.get_range(start, end)
        return result.__repr__()




import json

from pyraft.common import *
from pyraft.worker.worker import Worker
from pyraft.worker.worker import RespProtocol


# worker for data handling with redis interface
class RedisWorker(Worker):
    def __init__(self, addr):
        super().__init__(addr)
        self.init_redis_handler()
        self.set_protocol(RespProtocol())

    def init_redis_handler(self):
        # String
        self.handler['get'] = [self.do_get, 'r', 1, 1]
        self.handler['del'] = [self.do_del, 'we', 1, 1]
        self.handler['set'] = [self.do_set, 'we', 2, 2]
        self.handler['expire'] = [self.do_expire, 'we', 2, 2]
        self.handler['expireat'] = [self.do_expireat, 'we', 2, 2]
        self.handler['pexpire'] = [self.do_pexpire, 'we', 2, 2]
        self.handler['pexpireat'] = [self.do_pexpireat, 'we', 2, 2]

        # Hash
        self.handler['hgetall'] = [self.do_hgetall, 'r', 1, 1]
        self.handler['hset'] = [self.do_hset, 'we', 3, -1]
        self.handler['hget'] = [self.do_hget, 'r', 2, 2]
        self.handler['hdel'] = [self.do_hdel, 'we', 2, -1]
        self.handler['hlen'] = [self.do_hlen, 'r', 1, 1]

        # List
        self.handler['lpush'] = [self.do_lpush, 'we', 2, -1]
        self.handler['rpush'] = [self.do_rpush, 'we', 2, -1]
        self.handler['rpop'] = [self.do_rpop, 'we', 1, 1]
        self.handler['lpop'] = [self.do_lpop, 'we', 1, 1]
        self.handler['lrange'] = [self.do_lrange, 'r', 3, 3]
        self.handler['lindex'] = [self.do_lindex, 'r', 2, 2]
        self.handler['llen'] = [self.do_llen, 'r', 1, 1]
        self.handler['lset'] = [self.do_lset, 'we', 3, 3]
        self.handler['lrem'] = [self.do_lrem, 'we', 3, 3]
        self.handler['ltrim'] = [self.do_ltrim, 'we', 3, 3]

        # Tree (extended commands. not redis compatible)
        self.handler['tget'] = [self.do_tget, 'r', 1, 2] # tget path [attr]
        self.handler['tset'] = [self.do_tset, 'we', 3, -1] # tset path [json] | [attr data]*
        self.handler['tls'] = [self.do_tls, 'r', 1, 1] # tls path
        self.handler['tmk'] = [self.do_tmk, 'we', 1, 2] # tmk path [json] | [attr data]*
        self.handler['trm'] = [self.do_trm, 'we', 1, -1] # trm path [attr+]

    def do_tget(self, node, words):
        path = words[1].strip()

        nodes = path.split('/')
        pwd = node.data
        for n in nodes:
            pwd = pwd.get(n)
            if pwd is None:
                raise RaftException('node does not exists')

        if len(words) == 3:
            ret = pwd.get(words[2])
            if ret is None:
                raise RaftException('attr does not exists')

            return ret
        else:
            result = {}
            for k, v in pwd.items():
                if isinstance(v, dict):
                    continue

                result[k] = v

            return json.dumps(result)

    def do_tset(self, node, words):
        path = words[1].strip()

        if len(words) == 3:  # json
            kv_map = json.loads(words[2])
        else:
            kv_map = {}
            kv_list = words[2:]
            for i in range(0, len(kv_list), 2):
                kv_map[kv_list[i]] = kv_list[i + 1]

        nodes = path.split('/')
        pwd = node.data
        for n in nodes:
            pwd = pwd.get(n)
            if pwd is None:
                raise RaftException('node does not exists')

        for k, v in kv_map.items():
            if k in pwd and isinstance(k, dict):
                raise RaftException('can not overwrite node')
            pwd[k] = v

        return True

    def do_tls(self, node, words):
        path = words[1].strip()

        nodes = path.split('/')
        pwd = node.data
        for n in nodes:
            pwd = pwd.get(n)
            if pwd is None:
                raise RaftException('node does not exists')

        children = []
        for k, v in pwd.items():
            if isinstance(v, dict):
                children.append(k)

        return children

    def do_tmk(self, node, words):
        path = words[1].strip()

        if len(words) == 3: # json
            kv_map = json.loads(words[2])
        else:
            kv_map = {}
            kv_list = words[2:]
            for i in range(0, len(kv_list), 2):
                kv_map[kv_list[i]] = kv_list[i+1]

        nodes = path.split('/')
        pwd = node.data
        for n in nodes[:-1]:
            pwd = pwd.get(n)
            if pwd is None:
                raise RaftException('node does not exists')

        if nodes[-1] in pwd:
            raise RaftException('node or attr is already exists')

        pwd[nodes[-1]] = kv_map
        return True

    def do_trm(self, node, words):
        path = words[1].strip()
        attrs = words[2:]

        nodes = path.split('/')
        pwd = node.data
        parent = None
        for n in nodes:
            parent = pwd
            pwd = pwd.get(n)
            if pwd is None:
                raise RaftException('node does not exists')

        if len(attrs) > 0:
            for attr in attrs:
                del parent[attr]
        else:
            del parent[nodes[-1]]

        return True

    def do_lpush(self, node, words):
        key = words[1].strip()
        node.check_ttl(key)

        values = words[2:]

        if key not in node.data:
            node.data[key] = []

        lobj = node.data[key]
        if not isinstance(lobj, list):
            raise ERROR_TYPE

        for v in values:
            lobj.insert(0, v)

        return len(lobj)


    def do_rpush(self, node, words):
        key = words[1].strip()
        node.check_ttl(key)

        values = words[2:]

        if key not in node.data:
            node.data[key] = []

        lobj = node.data[key]

        if not isinstance(lobj, list):
            raise ERROR_TYPE

        for v in values:
            lobj.append(v)

        return len(lobj)

    def do_rpop(self, node, words):
        key = words[1].strip()
        node.check_ttl(key)

        if key in node.data:
            lobj = node.data[key]

            if not isinstance(lobj, list):
                raise ERROR_TYPE

            return lobj.pop()

        return None

    def do_lpop(self, node, words):
        key = words[1].strip()
        node.check_ttl(key)

        if key in node.data:
            lobj = node.data[key]

            if not isinstance(lobj, list):
                raise ERROR_TYPE

            return lobj.pop(0)

        return None

    def do_lrange(self, node, words):
        key = words[1].strip()
        node.check_ttl(key)

        start = int(words[2])
        end = int(words[3]) + 1

        if key in node.data:
            lobj = node.data[key]

            if not isinstance(lobj, list):
                raise ERROR_TYPE

            if end == 0:
                end = len(lobj)

            return lobj[start:end]

        return None

    def do_lindex(self, node, words):
        key = words[1].strip()
        node.check_ttl(key)

        index = int(words[2])

        if key in node.data:
            lobj = node.data[key]
            if not isinstance(lobj, list):
                raise ERROR_TYPE

            return lobj[index]

        return None

    def do_llen(self, node, words):
        key = words[1].strip()
        node.check_ttl(key)

        if key in node.data:
            lobj = node.data[key]
            if not isinstance(lobj, list):
                raise ERROR_TYPE

            return len(lobj)

        return 0

    def do_lset(self, node, words):
        key = words[1].strip()
        node.check_ttl(key)

        index = int(words[2])
        value = words[3]

        if key in node.data:
            lobj = node.data[key]

            if not isinstance(lobj, list):
                raise ERROR_TYPE

            lobj[index] = value

        return True

    def do_lrem(self, node, words):
        key = words[1].strip()
        node.check_ttl(key)

        count = int(words[2])
        value = words[3]

        if key in node.data:
            lobj = node.data[key]

            if not isinstance(lobj, list):
                raise ERROR_TYPE

            org_len = len(lobj)

            if count > 0:
                for i in range(count):
                    try:
                        lobj.remove(value)
                    except ValueError:
                        break

                return org_len - len(lobj)
            elif count == 0:
                l = [x for x in lobj if x != value]
                node.data[key] = l
                return org_len - len(l)
            else:
                raise ERROR_INVALID_PARAM

        return 0

    def do_ltrim(self, node, words):
        key = words[1].strip()
        node.check_ttl(key)

        start = int(words[2])
        end = int(words[3]) + 1

        if key in node.data:
            lobj = node.data[key]

            if not isinstance(lobj, list):
                raise ERROR_TYPE

            if end == 0:
                end = len(lobj)
            ret = lobj[start:end]
            node.data[key] = lobj

        return True

    def do_hgetall(self, node, words):
        key = words[1].strip()
        node.check_ttl(key)

        result = []
        if key in node.data:
            hobj = node.data[key]

            if not isinstance(hobj, dict):
                raise ERROR_TYPE

            for k, v in hobj.items():
                result.append(k)
                result.append(v)

        return result

    def do_hdel(self, node, words):
        key = words[1].strip()
        node.check_ttl(key)

        count = 0
        if key in node.data:
            hobj = node.data[key]

            if not isinstance(hobj, dict):
                raise ERROR_TYPE

            if words[2] in hobj:
                del (hobj[words[2]])
                count += 1

        return count

    def do_hset(self, node, words):
        key = words[1].strip()
        node.check_ttl(key)

        if key not in node.data:
            node.data[key] = {}

        hobj = node.data[key]
        if not isinstance(hobj, dict):
            raise ERROR_TYPE

        hobj[words[2]] = words[3]

        return True

    def do_hlen(self, node, words):
        key = words[1].strip()
        node.check_ttl(key)

        if key in node.data:
            hobj = node.data[key]
            if not isinstance(hobj, dict):
                raise ERROR_TYPE

            return len(hobj)

        return 0

    def do_hget(self, node, words):
        key = words[1].strip()
        node.check_ttl(key)

        if key in node.data:
            hobj = node.data[key]

            if not isinstance(hobj, dict):
                raise ERROR_TYPE

            if words[2] not in hobj:
                raise ERROR_NOT_EXISTS

            return hobj[words[2]]

        return None

    def do_get(self, node, words):
        key = words[1].strip()
        node.check_ttl(key)

        if key in node.data:
            value = node.data[key]
            if not isinstance(value, str):
                raise ERROR_TYPE

            return value

        return None

    def do_del(self, node, words):
        key = words[1].strip()
        if key in node.data:
            del node.data[key]
            node.clear_ttl(key)
            return 1

        return 0

    def do_set(self, node, words):
        node.data[words[1]] = words[2]
        node.clear_ttl(words[1])
        return True

    # TODO: cleanup expired by backgroud thread
    def pexpireat(self, node, key, ts):
        ts = float(ts) / 1000.0
        return node.set_ttl(key, ts)

    def do_pexpireat(self, node, words):
        ts = intcast(words[2])
        if ts == None:
            raise ERROR_CAST

        return self.pexpreat(node, words[1], ts)

    def do_pexpire(self, node, words):
        msec = intcast(words[2])
        if msec == None:
            raise ERROR_CAST

        return self.pexpireat(node, words[1], int(time.time() * 1000) + msec)

    def do_expireat(self, node, words):
        ts = intcast(words[2])
        if ts == None:
            raise ERROR_CAST

        return self.pexpreat(node, words[1], ts * 1000)

    def do_expire(self, node, words):
        sec = intcast(words[2])
        if sec == None:
            raise ERROR_CAST

        return self.pexpireat(node, words[1], int(time.time() * 1000) + (sec * 1000))




import threading, queue

from pyraft.common import *
from pyraft.protocol import resp

class LogItem(object):
    def __init__(self, term, index, ts, worker_offset, cmd):
        self.term = term
        self.index = index
        self.ts = ts
        self.cmd = cmd
        self.worker_offset = worker_offset

    def to_list(self):
        return [self.term, self.index, self.ts, self.worker_offset, self.cmd]

    def __repr__(self):
        return repr(self.to_list())


class LogFile(object):
    def __init__(self, nid, seq):
        self.seq = seq
        self.fn = 'raft_%s_%010d.log' % (nid, seq)
        self.fh = open(self.fn, 'w')
        self.count = 0
        self.start = 0
        self.end = 0

    def close(self):
        self.fh.close()

    def delete(self):
        self.fh.close()
        os.remove(self.fn)

    def append(self, item):
        self.fh.write(resp.resp_encoding(item.to_list()))
        self.fh.flush()
        self.count += 1

        if self.count == 0:
            self.start = item.index

        self.end = item.index

    def get_range(self, start, end=-1):
        decoded = []
        fh = open(self.fn, 'r')
        remain = fh.read()

        while True:
            l, remain = resp.resp_decoding(remain)
            if l == None:
                break

            decoded.append(LogItem(l[0], l[1], l[2], l[3], l[4]))

            if remain == '':
                break

        result = []
        for item in decoded:
            if item.index >= start:
                if end < 0 or item.index < end:
                    result.append(item)

        return result

    def start_index(self):
        return self.start

    def end_index(self):
        return self.end

    def size(self):
        return self.count


class LogFileStorage(object):
    def __init__(self, nid):
        self.index = 0
        self.file_list = []
        self.log = []
        self.nid = nid

    def append(self, item):
        self.index = item.index

        if len(self.file_list) == 0:
            lf = LogFile(self.nid, item.index)
            self.file_list.append(lf)
        else:
            count = self.file_list[-1].size() + 1
            if count > CONF_LOG_FILE_MAX:
                lf = LogFile(self.nid, item.index)
                self.file_list.append(lf)

        lf = self.file_list[-1]
        lf.append(item)

    def get_range(self, start, end=-1):
        result = []
        for item in self.file_list:
            if item.end_index() < start:
                continue

            if end > 0 and item.start_index() > end:
                break

            result += item.get_range(start, end)

        return result

    def cleanup(self, index):
        while len(self.file_list) > 0:
            f = self.file_list[0]
            if f.start_index() > index:
                break

            if f.end_index() <= index:
                f.delete()
                self.file_list.pop(0)
                continue

            break

    def start_index(self):
        if len(self.file_list) == 0:
            return self.index + 1

        return (self.file_list[0]).start_index()

    def size(self):
        count = 0

        for item in self.file_list:
            count += item.size()

        return count

    def close(self):
        for lf in self.file_list:
            lf.close()


class RaftLog(object):
    def __init__(self, nid):
        self.term = 0
        self.index = 0
        self.lock = threading.Lock()
        self.q = queue.Queue(4096)
        self.nid = nid
        self.log = LogFileStorage(self.nid)

        self.temp_item = []  # wait commit_index

    def apply_commit_index(self, commit_index):
        with self.lock:
            while len(self.temp_item) > 0:
                item = self.temp_item[0]
                if item == None:
                    del self.temp_item[0]
                    continue

                if item.index <= commit_index:
                    self._q_push(item)
                    last_index = item.index
                    del self.temp_item[0]
                else:
                    break

    def _q_push(self, item):
        self.q.put(item)

        if isinstance(item.cmd, Future):
            new_item = LogItem(item.term, item.index, item.ts, item.worker_offset, item.cmd.cmd)
            self.log.append(new_item)
        else:
            self.log.append(item)

    def push(self, item, commit_index):
        self.term = item.term
        self.index = item.index

        if item.index > commit_index:  # wait until commit
            with self.lock:
                for i in range(len(self.temp_item)):  # clean up invalid item
                    if self.temp_item[i].index >= item.index:
                        self.temp_item[i] = None

                self.temp_item.append(item)
            return

        with self.lock:
            self._q_push(item)

    def pop(self, timeout):
        try:
            item = self.q.get(True, timeout)
        except queue.Empty:
            return None

        return item

    def get_range(self, start, end=-1):
        with self.lock:
            return self.log.get_range(start, end)

    def cleanup(self, index):
        with self.lock:
            return self.log.cleanup(index)

    def start_index(self):
        with self.lock:
            return self.log.start_index()

    def size(self):
        with self.lock:
            return self.log.size()

    def get_index(self):
        return self.index

    def get_term(self):
        return self.term

    def close(self):
        self.log.close()






import threading

from pyraft.common import *
from pyraft.protocol import zk

class ZkEphermeralManager:
    def __init__(self, node):
        self.owner_path_node_map = {} # TODO: move to node.data (for restart handling)
        self.lock = threading.Lock()
        #self.session_timeout = 60.0
        self.session_timeout = 10.0
        self.node = node

    def start(self):
        self.quit_flag = False
        self.th_worker = threading.Thread(target=self.ephemeral_check)
        self.th_worker.start()

    def join(self):
        self.quit_flag = True
        self.th_worker.join()

    def ephemeral_check(self):
        while True:
            expire_list = []
            with self.lock:
                # change ownerping to hgetall
                zk_session = self.node.data.get('zk_session')
                if zk_session is None:
                    time.sleep(1.0)
                    continue

                owner_list = zk_session.keys()
                for owner in owner_list:
                    ts = zk_session.get(owner)
                    if ts is None: # can be removed while traversing
                        expire_list.append(owner)
                        continue

                    ts = int(ts)
                    #print('## check %s %d %d' % (owner, ts, time.time() - ts))
                    if time.time() - ts > self.session_timeout: # TODO: by timeout
                        expire_list.append(owner)

            for owner in expire_list:
                self.expire(owner)
                self.node.request('hdel', 'zk_session', str(owner))

            if self.quit_flag:
                return

            time.sleep(1.0)

    def regist(self, path, node):
        with self.lock:
            session = node.stat['ephemeralOwner']
            self.node.request_async('rpush', 'zk_sess_nodes_%s' % session, path)

    def unregist(self, path):
        with self.lock:
            node = self.worker._cd_path(self.node, path)
            session = node.stat['ephemeralOwner']
            self.node.request_async('lrem', 'zk_sess_nodes_%s' % session, path)

    def expire(self, session):
        with self.lock:
            logger.info('session %s is deleted' % session)
            print('## expire %s' % session)

            nodes = self.node.data.get('zk_sess_nodes_%s' % session)
            if nodes is None:
                return

            for node in nodes:
                logger.info('ephemeral node %s is deleted' % node)
                self.node.request('trm', 'ZK%s' % node)
            self.node.request('del', 'zk_sess_nodes_%s' % session)

            io = zk.get_session_io(session)
            if io is not None: # can be none if disconnected
                io.close()



import socket, select

from pyraft.common import *

class base_io:
    def __init__(self, sock):
        self.sock = sock
        self.buff = b''
        self.timeout = -1

        self.last_decodable = False
        self.last_buff_len = 0

    def connected(self):
        return self.sock != None

    def close(self):
        if self.sock != None:
            self.sock.close()

        self.sock = None

    def raw_write(self, msg):
        if self.sock == None:
            return None

        msg = self.raw_encode(msg)

        try:
            if isinstance(msg, str):
                msg = msg.encode()
            ret = self.sock.send(msg)
        except socket.error:
            self.close()
            return None

        if ret == 0:
            self.close()
            return None

        return ret

    def write(self, msg):
        if self.sock == None:
            return None

        try:
            msg = self.encode(msg)
            if isinstance(msg, str):
                msg = msg.encode()

            ret = self.sock.send(msg)
            if ret == 0:
                self.close()
                return None
            return ret

        except socket.error:
            self.close()
            return None

    def read(self, timeout=None):
        if self.sock == None:
            return None

        while True:
            readable = False
            if timeout != None: # avoid useless decodable check
                if self.last_decodable == False and self.last_buff_len == len(self.buff):
                    reads, writes, excepts = select.select([self.sock], [], [], timeout)
                    if len(reads) == 0:
                        return b''
                    else:
                        readable = True

            self.last_decodable = True
            if not self.decodable(self.buff):
                if timeout != None and not readable: # skip double wait
                    reads, writes, excepts = select.select([self.sock], [], [], timeout)
                    if len(reads) == 0:
                        self.last_decodable = False
                        self.last_buff_len = len(self.buff)
                        return b''

                try:
                    tmp = self.sock.recv(4096)
                except socket.error:
                    self.close()
                    return None
                except Exception:
                    self.close()
                    return None

                if tmp == b'':
                    self.close()
                    return None

                self.buff += tmp

            result, self.buff = self.decode(self.buff)
            if result == None:
                if timeout != None:
                    reads, writes, excepts = select.select([self.sock], [], [], timeout)
                    if len(reads) == 0:
                        self.last_decodable = False
                        self.last_buff_len = len(self.buff)
                        return b''

                try:
                    tmp = self.sock.recv(4096)
                except socket.error:
                    self.close()
                    return None

                if tmp == b'':
                    self.close()
                    return None

                self.buff += tmp
                continue

            return result

    def read_all(self, timeout=None):
        result = []

        while True:
            ret = self.read(timeout)
            if ret == None:
                return None

            if ret == b'':
                break

            result.append(ret)
            item, remain = self.decode(self.buff)
            if item == None:
                break

        return result

    # inherit below 4 interface
    def raw_encode(self, msg):
        pass

    def encode(self, msg):
        pass

    def decode(self, msg):
        pass

    def decodable(self, buff):
        pass




import traceback

from pyraft.common import *
from pyraft.protocol import resp


class RespProtocol:
    def open_io(self, handle):
        return resp.resp_io(handle)


class Worker:
    def __init__(self, addr):
        self.handler = {}
        self.shutdown_flag = False
        self.p = None

        self.addr = addr
        self.ip, self.port = addr.split(':')
        self.port = int(self.port)
        self.worker_offset = 0 # offset between node baseport

    def init_node(self, node):
        pass # inherit for worker specific node init

    def set_protocol(self, p):
        self.p = p

    def start(self, node):
        if self.p is None:
            self.p = RespProtocol()

        self.th_worker = threading.Thread(target=self.worker_listen, args=(node,))
        self.th_worker.start()

    def shutdown(self):
        self.shutdown_flag = True

    def join(self):
        self.th_worker.join()

    def worker_listen(self, node):
        self.worker_listen_sock = socket.socket()
        self.worker_listen_sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.worker_listen_sock.bind((self.ip, self.port))
        self.worker_listen_sock.listen(1)
        self.worker_listen_sock.settimeout(1)

        while True:
            try:
                sock, addr = self.worker_listen_sock.accept()
                wt = threading.Thread(target=self.process_work, args=(node, sock,))
                wt.start()
            except socket.timeout:
                if self.shutdown_flag:
                    self.worker_listen_sock.close()
                    break

    def get_handler(self, name):
        if name not in self.handler:
            return None

        return self.handler[name]

    def process_work(self, node, sock):
        pio = self.p.open_io(sock)

        while True:
            if self.shutdown_flag:
                pio.close()
                return

            words = pio.read(timeout=1.0)
            if words == None:
                node.log_info('disconnected')
                pio.close()
                return

            if isinstance(words, str): # resp inline command
                words = words.split()

            if words == b'' or words == ['']:
                continue

            if len(words) > 0:
                handler = self.get_handler(words[0].lower())
                if handler == None:
                    pio.write(Exception('Unknown command: %s' % words[0]))
                    continue

                p_min = handler[2]
                p_max = handler[3]
                if len(words) - 1 < p_min:
                    pio.write(Exception('insufficient param'))
                    continue

                if p_max > 0 and len(words) - 1 > p_max:
                    pio.write(Exception('too many param'))
                    continue

                try:
                    ret = node.propose(words, self.worker_offset)
                except RaftException as e:
                    ret = e
                except Exception as e:
                    print('unexpected exception: ', traceback.format_exc())
                    ret = e

                # for special actions (like quit)
                if isinstance(ret, dict):
                    if 'bypass' in ret:
                        pio.raw_write(ret['bypass'])
                        continue

                    if 'quit' in ret:
                        pio.close()
                        return

                    # some protocols have to send something before closing
                    if 'quit_after_send' in ret:
                        pio.write(ret['quit_after_send'])
                        pio.close()
                        return

                pio.write(ret)

    def relay_cmd(self, leader, cmd, worker_offset):
        p = leader
        try:
            if not hasattr(p, 'req_io'):
                ip, port = p.addr.split(':')
                port = int(port) + worker_offset
                sock = socket.socket()
                sock.connect((ip, port))
                p.req_io = self.p.open_io(sock)

            p.req_io.write(cmd)
            return p.req_io.read()

        except Exception as e:
            p.req_io.close()
            delattr(p, 'req_io')
            raise RaftException('relay to leader has exception: %s' % str(e))


class MergedWorker(Worker):
    def __init__(self, addr, *workers):
        super().__init__(addr)
        self.worker_list = workers

    def get_handler(self, name):
        if name in self.handler:
            return self.handler[name]

        for worker in self.worker_list:
            handler = worker.get_handler(name)
            if handler is not None:
                return handler

        return None

    def append_worker(self, worker):
        self.worker_list.append(worker)



# coding: utf-8

import setuptools
from setuptools import setup
from os import path

__version__ = '1.0.0'

requirements = [req.strip() for req in open('requirements.txt').readlines()]

short_description ='python raft implementation with resp interface'

try:
	import pypandoc
	long_description = pypandoc.convert_file('README.md', 'rst')
except(IOError, ImportError):
	long_description = open('README.md').read()

setup(
    name='pyraft',
    packages=setuptools.find_packages(),
    version=__version__,
    description=short_description,
    long_description=long_description,
    author='Lee, Ki-Yeul',
    author_email='lynix94@gmail.com',
    license='MIT',
    url='https://github.com/lynix94/pyraft',
    keywords=['python', 'raft', 'replication', 'pyraft'],
    install_requires=requirements,
	python_requires  = '>=3',
    classifiers = [
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ]
)



import os, sys, time
import socket, threading, logging

CONF_LOG_MAX = 100000
CONF_LOG_FILE_MAX = 10000

CONF_VOTING_TIME = 1.0
CONF_PING_TIMEOUT = 5  # re-elect leader after CONF_PING_TIMEOUT

def intcast(src):
    if isinstance(src, int):
        return src

    if src.isdigit() == False:
        return None

    return int(src)


ERROR_CAST = Exception('number format error')
ERROR_APPEND_ENTRY = Exception('append entry failed')
ERROR_TYPE = Exception('invalid data type')
ERROR_NOT_EXISTS = Exception('not exists')
ERROR_INVALID_PARAM = Exception('invalid parameter')

class RaftException(Exception):
    def __init__(self, msg):
        super().__init__(msg)

class Future(object):
    def __init__(self, cmd, worker_offset=0):
        self.worker_offset = worker_offset
        self.cmd = cmd
        self.value = None
        self.cond = threading.Condition()

    def get(self, timeout=None):
        if self.value != None:
            return self.value

        try:
            with self.cond:
                self.cond.wait(timeout)
        except RuntimeError:
            return None

        return self.value

    def set(self, value):
        with self.cond:
            self.value = value
            self.cond.notify()

def bytes_to_str(b):
    out = []
    for c in b:
        out.append('%02x' % c)

    return ' '.join(out)

logger = logging.getLogger('pyraft')


# basic proxy tool for protocol analysis

import socket, threading, select
import argparse

def dump_str(data):
	result = []
	result_ascii = []
	for i in data:
		result.append('%02x ' % i)
		result_ascii.append('%c ' % i)

	return '%s - %s' % (''.join(result), ''.join(result_ascii))

def handle_client(server, c, addr):
	with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
		ip, port = server.split(':')
		s.connect((ip, int(port)))

		while True:
			rl, wl, el = select.select([c], [], [])
			for r in rl:
				data = r.recv(4096)
				if not data:
					c.close()
					return

				print ('<< req[%d]: %s' % (len(data), dump_str(data)))
				s.sendall(data)

			rl, wl, el = select.select([s], [], [], 0.01)
			for r in rl:
				data = r.recv(4096)
				if not data:
					c.close()
					return

				print ('>> resp[%d]: %s' % (len(data), dump_str(data)))
				c.sendall(data)



def do_proxy(server, listen):
	s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
	ip, port = listen.split(':')
	s.bind((ip, int(port)))
	s.listen(10)
	while True:
		conn, addr = s.accept()
		th = threading.Thread(target=handle_client, args=(server, conn, addr))
		th.start()

if __name__ == '__main__':
	parser = argparse.ArgumentParser()
	parser.add_argument('-s', dest='server', help='server side address', required=True)
	parser.add_argument('-a', dest='addr', help='listen address', required=True)

	args = parser.parse_args()

	do_proxy(args.server, args.addr)

	



from pyraft.common import RaftException
from collections import defaultdict

# referred and copied from kazoo exceptions

class ZkException(RaftException): # internal exception (not for response)
    def __init__(self, msg):
        super().__init__(msg)

class ZookeeperError(RaftException):
    def __init__(self):
        super().__init__('')

def _invalid_error_code():
    raise Exception('invalid error code')

EXCEPTIONS = defaultdict(_invalid_error_code)

def _zookeeper_exception(code):
    def decorator(klass):
        EXCEPTIONS[code] = klass
        klass.code = code
        return klass

    return decorator

@_zookeeper_exception(0)
class RolledBackError(ZookeeperError):
    pass

@_zookeeper_exception(-1)
class SystemZookeeperError(ZookeeperError):
    pass

@_zookeeper_exception(-2)
class RuntimeInconsistency(ZookeeperError):
    pass

@_zookeeper_exception(-3)
class DataInconsistency(ZookeeperError):
    pass

@_zookeeper_exception(-4)
class ConnectionLoss(ZookeeperError):
    pass

@_zookeeper_exception(-5)
class MarshallingError(ZookeeperError):
    pass

@_zookeeper_exception(-6)
class UnimplementedError(ZookeeperError):
    pass

@_zookeeper_exception(-7)
class OperationTimeoutError(ZookeeperError):
    pass

@_zookeeper_exception(-8)
class BadArgumentsError(ZookeeperError):
    pass

@_zookeeper_exception(-13)
class NewConfigNoQuorumError(ZookeeperError):
    pass

@_zookeeper_exception(-14)
class ReconfigInProcessError(ZookeeperError):
    pass

@_zookeeper_exception(-100)
class APIError(ZookeeperError):
    pass

@_zookeeper_exception(-101)
class NoNodeError(ZookeeperError):
    pass

@_zookeeper_exception(-102)
class NoAuthError(ZookeeperError):
    pass

@_zookeeper_exception(-103)
class BadVersionError(ZookeeperError):
    pass

@_zookeeper_exception(-108)
class NoChildrenForEphemeralsError(ZookeeperError):
    pass

@_zookeeper_exception(-110)
class NodeExistsError(ZookeeperError):
    pass

@_zookeeper_exception(-111)
class NotEmptyError(ZookeeperError):
    pass

@_zookeeper_exception(-112)
class SessionExpiredError(ZookeeperError):
    pass

@_zookeeper_exception(-113)
class InvalidCallbackError(ZookeeperError):
    pass

@_zookeeper_exception(-114)
class InvalidACLError(ZookeeperError):
    pass

@_zookeeper_exception(-115)
class AuthFailedError(ZookeeperError):
    pass

@_zookeeper_exception(-118)
class SessionMovedError(ZookeeperError):
    pass

@_zookeeper_exception(-119)
class NotReadOnlyCallError(ZookeeperError):
    """An API call that is not read-only was used while connected to
    a read-only server"""

class ConnectionClosedError(SessionExpiredError):
    """Connection is closed"""

